# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains injection functionality for the cell."""
from __future__ import annotations
import math
import warnings
import logging

import neuron
import numpy as np
from typing_extensions import deprecated

from bluecellulab.cell.stimuli_generator import (
    gen_ornstein_uhlenbeck,
    gen_shotnoise_signal,
    get_relative_shotnoise_params,
)
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.rngsettings import RNGSettings
from bluecellulab.stimulus.circuit_stimulus_definitions import (
    ClampMode,
    Hyperpolarizing,
    Noise,
    OrnsteinUhlenbeck,
    ShotNoise,
    RelativeOrnsteinUhlenbeck,
    RelativeShotNoise,
)
from bluecellulab.type_aliases import NeuronSection, TStim


logger = logging.getLogger(__name__)


class InjectableMixin:
    """Mixin responsible of injections to the cell.

    Important Usage Note: Adds the instantiated Neuron objects to
     self.persistent to explicitly destroy them when their lifetime ends.
    """

    def relativity_proportion(self, stim_mode: ClampMode) -> float:
        """Relativity proportion used in Relative stimuli e.g. relative
        shotnoise."""
        if stim_mode == ClampMode.CONDUCTANCE:
            input_resistance = self.sonata_proxy.get_input_resistance().iloc[0]  # type: ignore
            return 1.0 / input_resistance
        else:  # Current
            return self.threshold  # type: ignore

    def add_pulse(self, stimulus) -> TStim:
        """Inject pulse stimulus for replay."""
        tstim = neuron.h.TStim(0.5, sec=self.soma)  # type: ignore
        tstim.train(stimulus.delay,
                    stimulus.duration,
                    stimulus.amp_start,
                    stimulus.frequency,
                    stimulus.width)
        self.persistent.append(tstim)  # type: ignore
        return tstim

    def add_step(
        self,
        start_time: float,
        stop_time: float,
        level: float,
        section: NeuronSection | None = None,
        segx: float = 0.5,
    ) -> TStim:
        """Add a step current injection.

        Args:
            start_time: Start time of the step injection in seconds.
            stop_time: Stop time of the step injection in seconds.
            level: Current level to inject in nanoamperes (nA).
            section: The section to inject current into.
                Defaults to the soma section.
            segx: The fractional location within the section to inject.
                Defaults to 0.5 (center of the section).

        Returns:
            TStim NEURON object responsible of vectoral injection.
        """
        if section is None:
            section = self.soma  # type: ignore

        tstim = neuron.h.TStim(segx, sec=section)
        duration = stop_time - start_time
        tstim.pulse(start_time, duration, level)
        self.persistent.append(tstim)  # type: ignore
        return tstim

    def add_ramp(
        self,
        start_time: float,
        stop_time: float,
        start_level: float,
        stop_level: float,
        section: NeuronSection | None = None,
        segx: float = 0.5,
    ) -> TStim:
        """Add a ramp current injection.

        Args:
            start_time: Start time of the ramp injection in seconds.
            stop_time: Stop time of the ramp injection in seconds.
            start_level: Current level at the start of the ramp in nanoamperes (nA).
            stop_level: Current level at the end of the ramp in nanoamperes (nA).
            section: The section to inject current into (optional). Defaults to soma.
            segx: The fractional location within the section to inject (optional).

        Returns:
            TStim NEURON object responsible of vectoral injection.
        """
        if section is None:
            section = self.soma  # type: ignore
        tstim = neuron.h.TStim(segx, sec=section)
        tstim.ramp(
            0.0,
            start_time,
            start_level,
            stop_level,
            stop_time - start_time,
            0.0,
            0.0)
        self.persistent.append(tstim)  # type: ignore
        return tstim

    def add_voltage_clamp(
            self, stop_time, level, rs=None, section=None, segx=0.5,
            current_record_name=None, current_record_dt=None):
        """Add a voltage clamp.

        Parameters
        ----------

        stop_time : float
            Time at which voltage clamp should stop
        level : float
            Voltage level of the vc (in mV)
        rs: float
            Series resistance of the vc (in MOhm)
        section: NEURON object
            Object representing the section to place the vc
        segx: float
            Segment x coordinate to place the vc
        current_record_name: str
            Name of the recording that will store the current
        current_record_dt: float
            Timestep to use for the recording of the current

        Returns
        -------

        SEClamp (NEURON) object of the created vc
        """

        if section is None:
            section = self.soma
        if current_record_dt is None:
            current_record_dt = self.record_dt
        vclamp = neuron.h.SEClamp(segx, sec=section)
        self.persistent.append(vclamp)

        vclamp.amp1 = level
        vclamp.dur1 = stop_time

        if rs is not None:
            vclamp.rs = rs

        current = neuron.h.Vector()
        if current_record_dt is None:
            current.record(vclamp._ref_i)
        else:
            current.record(
                vclamp._ref_i,
                self.get_precise_record_dt(current_record_dt))

        self.recordings[current_record_name] = current
        return vclamp

    def _get_noise_step_rand(self, noisestim_count):
        """Return rng for noise step stimulus."""
        rng_settings = RNGSettings.get_instance()
        if rng_settings.mode == "Compatibility":
            rng = neuron.h.Random(self.cell_id.id + noisestim_count)
        elif rng_settings.mode == "UpdatedMCell":
            rng = neuron.h.Random()
            rng.MCellRan4(
                noisestim_count * 10000 + 100,
                rng_settings.base_seed +
                rng_settings.stimulus_seed +
                self.cell_id.id * 1000)
        elif rng_settings.mode == "Random123":
            rng = neuron.h.Random()
            rng.Random123(
                noisestim_count + 100,
                rng_settings.stimulus_seed + 500,
                self.cell_id.id + 300)

        self.persistent.append(rng)
        return rng

    def add_noise_step(self, section,
                       segx,
                       mean, variance,
                       delay,
                       duration, seed=None, noisestim_count=0):
        """Inject a step current with noise on top."""
        if seed is not None:
            rand = neuron.h.Random(seed)
        else:
            rand = self._get_noise_step_rand(noisestim_count)

        tstim = neuron.h.TStim(segx, rand, sec=section)
        tstim.noise(delay, duration, mean, variance)
        self.persistent.append(rand)
        self.persistent.append(tstim)
        return tstim

    def add_replay_noise(
            self,
            stimulus: Noise,
            noise_seed=None,
            noisestim_count=0):
        """Add a replay noise stimulus."""
        mean = (stimulus.mean_percent * self.threshold) / 100.0  # type: ignore
        variance = (stimulus.variance * self.threshold) / 100.0  # type: ignore
        tstim = self.add_noise_step(
            self.soma,  # type: ignore
            0.5,
            mean,
            variance,
            stimulus.delay,
            stimulus.duration,
            seed=noise_seed,
            noisestim_count=noisestim_count)

        return tstim

    def add_replay_hypamp(self, stimulus: Hyperpolarizing):
        """Inject hypamp for the replay."""
        tstim = neuron.h.TStim(0.5, sec=self.soma)  # type: ignore
        if self.hypamp is None:  # type: ignore
            raise BluecellulabError("Cell.hypamp must be set for hypamp stimulus")
        amp: float = self.hypamp  # type: ignore
        tstim.pulse(stimulus.delay, stimulus.duration, amp)
        self.persistent.append(tstim)  # type: ignore
        return tstim

    def add_replay_relativelinear(self, stimulus):
        """Add a relative linear stimulus."""
        tstim = neuron.h.TStim(0.5, sec=self.soma)
        amp = stimulus.percent_start / 100.0 * self.threshold
        tstim.pulse(stimulus.delay, stimulus.duration, amp)
        self.persistent.append(tstim)

        return tstim

    def _get_ornstein_uhlenbeck_rand(self, stim_count, seed):
        """Return rng for ornstein_uhlenbeck simulation."""
        rng_settings = RNGSettings.get_instance()
        if rng_settings.mode == "Random123":
            seed1 = stim_count + 2997  # stimulus block
            seed2 = rng_settings.stimulus_seed + 291204  # stimulus type
            seed3 = self.cell_id.id + 123 if seed is None else seed  # GID
            logger.debug("Using ornstein_uhlenbeck process seeds %d %d %d" %
                         (seed1, seed2, seed3))
            rng = neuron.h.Random()
            rng.Random123(seed1, seed2, seed3)
        else:
            raise BluecellulabError("Shot noise stimulus requires Random123")

        self.persistent.append(rng)
        return rng

    def _get_shotnoise_step_rand(self, shotnoise_stim_count, seed=None):
        """Return rng for shot noise step stimulus."""
        rng_settings = RNGSettings.get_instance()
        if rng_settings.mode == "Random123":
            seed1 = shotnoise_stim_count + 2997
            seed2 = rng_settings.stimulus_seed + 19216
            seed3 = self.cell_id.id + 123 if seed is None else seed
            logger.debug("Using shot noise seeds %d %d %d" %
                         (seed1, seed2, seed3))
            rng = neuron.h.Random()
            rng.Random123(seed1, seed2, seed3)
        else:
            raise BluecellulabError("Shot noise stimulus requires Random123")

        self.persistent.append(rng)
        return rng

    def inject_current_clamp_signal(self, section, segx, tvec, svec):
        """Inject any signal via current clamp."""
        cs = neuron.h.IClamp(segx, sec=section)
        cs.dur = tvec[-1]
        svec.play(cs._ref_amp, tvec, 1)

        self.persistent.append(cs)
        self.persistent.append(tvec)
        self.persistent.append(svec)
        return tvec, svec

    def inject_dynamic_clamp_signal(self, section, segx, tvec, svec, reversal):
        """Injects any signal via a dynamic conductance clamp.

        Args:
            reversal (float): reversal potential of conductance (mV)
        """
        clamp = neuron.h.SEClamp(segx, sec=section)
        clamp.dur1 = tvec[-1]
        clamp.amp1 = reversal
        # support delay with initial zero
        tvec.insrt(0, 0)
        svec.insrt(0, 0)
        # replace svec with inverted and clamped signal
        # rs is in MOhm, so conductance is in uS (micro Siemens)
        svec = neuron.h.Vector(
            [1 / x if x > 1E-9 and x < 1E9 else 1E9 for x in svec])
        svec.play(clamp._ref_rs, tvec, 1)

        self.persistent.append(clamp)
        self.persistent.append(tvec)
        self.persistent.append(svec)
        return tvec, svec

    def add_replay_shotnoise(
            self,
            section,
            segx,
            stimulus: ShotNoise,
            shotnoise_stim_count=0):
        """Add a replay shot noise stimulus."""
        rng = self._get_shotnoise_step_rand(shotnoise_stim_count, stimulus.seed)
        tvec, svec = gen_shotnoise_signal(stimulus.decay_time, stimulus.rise_time, stimulus.rate, stimulus.amp_mean,
                                          stimulus.amp_var, stimulus.duration, stimulus.dt, rng=rng)
        tvec.add(stimulus.delay)  # add delay

        if stimulus.mode == ClampMode.CONDUCTANCE:
            return self.inject_dynamic_clamp_signal(section, segx, tvec, svec, stimulus.reversal)
        else:
            return self.inject_current_clamp_signal(section, segx, tvec, svec)

    def add_replay_relative_shotnoise(
            self,
            section,
            segx,
            stimulus: RelativeShotNoise,
            shotnoise_stim_count=0):
        """Add a replay relative shot noise stimulus."""
        cv_square = stimulus.amp_cv**2

        stim_mode = stimulus.mode
        rel_prop = self.relativity_proportion(stim_mode)

        mean = stimulus.mean_percent / 100 * rel_prop
        sd = stimulus.sd_percent / 100 * rel_prop
        var = sd * sd

        rate, amp_mean, amp_var = get_relative_shotnoise_params(
            mean, var, stimulus.decay_time, stimulus.rise_time, cv_square)

        rng = self._get_shotnoise_step_rand(shotnoise_stim_count, stimulus.seed)
        tvec, svec = gen_shotnoise_signal(stimulus.decay_time, stimulus.rise_time, rate, amp_mean,
                                          amp_var, stimulus.duration, stimulus.dt, rng=rng)
        tvec.add(stimulus.delay)  # add delay

        if stim_mode == ClampMode.CONDUCTANCE:
            return self.inject_dynamic_clamp_signal(section, segx, tvec, svec, stimulus.reversal)
        else:
            return self.inject_current_clamp_signal(section, segx, tvec, svec)

    def add_ornstein_uhlenbeck(
        self, section, segx, stimulus: OrnsteinUhlenbeck, stim_count=0
    ):
        """Add an Ornstein-Uhlenbeck process, injected as current or
        conductance."""
        rng = self._get_ornstein_uhlenbeck_rand(stim_count, stimulus.seed)
        tvec, svec = gen_ornstein_uhlenbeck(
            stimulus.tau,
            stimulus.sigma,
            stimulus.mean,
            stimulus.duration,
            stimulus.dt,
            rng,
        )

        tvec.add(stimulus.delay)  # add delay

        if stimulus.mode == ClampMode.CONDUCTANCE:
            return self.inject_dynamic_clamp_signal(
                section, segx, tvec, svec, stimulus.reversal
            )
        else:
            return self.inject_current_clamp_signal(section, segx, tvec, svec)

    def add_relative_ornstein_uhlenbeck(
        self, section, segx, stimulus: RelativeOrnsteinUhlenbeck, stim_count=0
    ):
        """Add an Ornstein-Uhlenbeck process, injected as current or
        conductance, relative to cell threshold current or inverse input
        resistance."""
        stim_mode = stimulus.mode
        rel_prop = self.relativity_proportion(stim_mode)

        sigma = stimulus.sd_percent / 100 * rel_prop
        if sigma <= 0:
            raise BluecellulabError(f"standard deviation: {sigma}, must be positive.")

        mean = stimulus.mean_percent / 100 * rel_prop
        if mean < 0 and abs(mean) > 2 * sigma:
            warnings.warn("relative ornstein uhlenbeck signal is mostly zero.")

        rng = self._get_ornstein_uhlenbeck_rand(stim_count, stimulus.seed)
        tvec, svec = gen_ornstein_uhlenbeck(
            stimulus.tau, sigma, mean, stimulus.duration, stimulus.dt, rng
        )

        tvec.add(stimulus.delay)  # add delay

        if stim_mode == ClampMode.CONDUCTANCE:
            return self.inject_dynamic_clamp_signal(section, segx, tvec, svec, stimulus.reversal)
        else:
            return self.inject_current_clamp_signal(section, segx, tvec, svec)

    def inject_current_waveform(self, t_content, i_content, section=None, segx=0.5):
        """Inject a custom current waveform into the cell."""
        if section is None:
            section = self.soma

        time_vector = neuron.h.Vector().from_python(t_content)
        current_vector = neuron.h.Vector().from_python(i_content)

        iclamp = neuron.h.IClamp(segx, sec=section)
        self.persistent.extend([iclamp, time_vector, current_vector])

        iclamp.delay = t_content[0]
        iclamp.dur = t_content[-1] - t_content[0]
        current_vector.play(iclamp._ref_amp, time_vector)

        return iclamp, current_vector

    @deprecated("Use add_sin_current instead.")
    def addSineCurrentInject(self, start_time, stop_time, freq,
                             amplitude, mid_level, dt=1.0):
        """Add a sinusoidal current injection.

        Returns
        -------

        (numpy array, numpy array) : time and current data
        """
        t_content = np.arange(start_time, stop_time, dt)
        i_content = [amplitude * math.sin(freq * (x - start_time) * (
            2 * math.pi)) + mid_level for x in t_content]
        self.inject_current_waveform(t_content, i_content)
        return (t_content, i_content)

    def add_sin_current(self, amp, start_time, duration, frequency,
                        section=None, segx=0.5):
        """Add a sinusoidal current to the cell."""
        if section is None:
            section = self.soma
        tstim = neuron.h.TStim(segx, sec=section)
        tstim.sin(amp, start_time, duration, frequency)
        self.persistent.append(tstim)
        return tstim

    def add_alpha_synapse(
        self,
        onset: float,
        tau: float,
        gmax: float,
        e: float,
        section: NeuronSection,
        segx=0.5,
    ) -> NeuronSection:
        """Add an AlphaSynapse NEURON point process stimulus to the cell."""
        syn = neuron.h.AlphaSynapse(segx, sec=section)
        syn.onset = onset
        syn.tau = tau
        syn.gmax = gmax
        syn.e = e
        self.persistent.append(syn)  # type: ignore
        return syn



"""Functions for configuring the verbosity of BlueCelluLab."""


import logging
import os
from typing import Optional

import bluecellulab


logger = logging.getLogger(__name__)

VERBOSE_LEVEL = 0
ENV_VERBOSE_LEVEL: Optional[str] = None


def set_verbose(level: int = 1) -> None:
    """Set the verbose level of BluecellulabError.

    Parameters
    ----------
    level :
            Verbose level, the higher the more verbosity.
            Level 0 means 'completely quiet', except if some very serious
            errors or warnings are encountered.
    """
    bluecellulab.VERBOSE_LEVEL = level

    if level <= 0:
        logging.getLogger('bluecellulab').setLevel(logging.CRITICAL)
    elif level == 1:
        logging.getLogger('bluecellulab').setLevel(logging.ERROR)
    elif level == 2:
        logging.getLogger('bluecellulab').setLevel(logging.WARNING)
    elif level > 2 and level <= 5:
        logging.getLogger('bluecellulab').setLevel(logging.INFO)
    else:
        logging.getLogger('bluecellulab').setLevel(logging.DEBUG)


def set_verbose_from_env() -> None:
    """Get verbose level from environment."""
    bluecellulab.ENV_VERBOSE_LEVEL = os.environ.get('BLUECELLULAB_VERBOSE_LEVEL')

    if bluecellulab.ENV_VERBOSE_LEVEL is not None:
        set_verbose(int(bluecellulab.ENV_VERBOSE_LEVEL))


set_verbose_from_env()



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that represents a dendrogram window."""
from __future__ import annotations
import numpy as np
import pylab

from bluecellulab.psection import PSection
from bluecellulab.psegment import PSegment


def setup_draw(psegments: list[PSegment], maxsegdiam: float, figure, x, y, variable=None, varbounds=None) -> None:
    """Setup draw of psection."""
    y_accum = 0.0
    for psegment in psegments:
        psegment.setupDraw(figure,
                           x + (maxsegdiam - psegment.diam) / 2,
                           y + y_accum,
                           variable=variable,
                           varbounds=varbounds)
        y_accum += psegment.L


def draw_tree(psection: PSection, figure, x, y, variable=None, varbounds=None) -> None:
    """Draw a dendritic tree."""
    # Draw myself
    setup_draw(
        psection.psegments, psection.maxsegdiam, figure, x, y, variable=variable, varbounds=varbounds
    )

    # Draw children

    # First child is a same x coordinate
    new_x = x  # + self.L + self.xSpacing

    # Children drawn L + ySpacing heigher
    new_y = y + psection.L + psection.ySpacing

    for child in psection.pchildren:
        draw_tree(child, figure, new_x, new_y, variable=variable, varbounds=varbounds)
        pylab.plot(
            [x + psection.diam / 2, new_x + child.diam / 2],
            [y + psection.L, new_y], 'k')
        # Prepare new_x for next child
        new_x = new_x + child.tree_width()


def redraw_psection(psection: PSection) -> None:
    """Redraw psection."""
    for psegment in psection.psegments:
        psegment.redraw()


class Dendrogram:
    """Class that represent a dendrogram plot."""

    def __init__(
            self,
            psections: list[PSection],
            variable=None,
            active=False,
            save_fig_path=None,
            interactive=False,
            scale_bar=True,
            scale_bar_size=10.0,
            fig_title=None,
            fig_show=True):

        if interactive:
            pylab.ion()

        self.dend_figure = pylab.figure(figsize=(20, 12))

        title_space = 0.0
        if fig_title:
            title_space = 30.0
            self.dend_figure.suptitle(fig_title)

        if interactive:
            pylab.ioff()

        self.psections = psections
        # neuron.h.finitialize()

        self.proot: PSection = psections[0]
        self.psections = [self.proot] + self.proot.all_descendants()

        xSpacing = self.proot.xSpacing
        ySpacing = self.proot.ySpacing

        max_y = self.proot.tree_height() + self.proot.ySpacing + title_space
        max_x = self.proot.tree_width() + self.proot.xSpacing + scale_bar_size
        pylab.xlim([0, max_x])
        pylab.ylim([0, max_y])
        pylab.gca().set_xticks([])
        pylab.gca().set_yticks([])
        pylab.gcf().subplots_adjust(
            top=0.99, bottom=0.01, left=0.01, right=0.99, hspace=0.3)

        if variable == "v" or variable is None:
            varbounds = [-100, 50]
        else:
            varbounds = self.proot.getTreeVarBounds(variable)

        if variable is not None:
            cax = pylab.imshow(np.outer(np.arange(0, 1, 0.1), np.ones(
                1)), aspect='auto', cmap=pylab.get_cmap("hot"), origin="lower")
            pylab.clim(varbounds[0], varbounds[1])

            cbar = self.dend_figure.colorbar(
                cax, ticks=[varbounds[0], varbounds[1]])
            cbar.ax.set_yticklabels(["%.2e" % (
                varbounds[0]), "%.2e" % (varbounds[1])])

        draw_tree(self.proot, self.dend_figure, self.proot.xSpacing,
                  self.proot.ySpacing, variable=variable,
                  varbounds=varbounds)

        if scale_bar:
            pylab.plot(
                [max_x - xSpacing, max_x - xSpacing, max_x - xSpacing, max_x -
                 xSpacing - scale_bar_size],
                [ySpacing, ySpacing + scale_bar_size, ySpacing, ySpacing], 'k',
                linewidth=2)
            pylab.text(
                max_x -
                xSpacing -
                0.9 * scale_bar_size,
                2 *
                ySpacing,
                "10 micron",
                fontsize=8)

            color_types = [
                ('m', 'apical'),
                ('k', 'soma'),
                ('b', 'AIS'),
                ('r', 'basal'),
                ('g', 'myelin')
            ]
            for i, (color, section_type) in enumerate(color_types):
                pylab.text(
                    max_x -
                    xSpacing -
                    0.9 *
                    scale_bar_size,
                    (i + 2) * 3 * scale_bar_size,
                    section_type,
                    fontsize=8,
                    bbox=dict(
                        facecolor=color, alpha=0.5))

        self.dend_figure.canvas.draw()

        for section in self.psections:
            section_id = section.isec
            if section_id is not None:
                redraw_psection(psections[section_id])

        self.canvas = self.dend_figure.gca().figure.canvas
        self.ax = self.dend_figure.gca()
        self.canvas.blit(self.ax.bbox)

        self.background = self.canvas.copy_from_bbox(
            self.dend_figure.gca().bbox)
        self.drawCount = 1

        self.active = active

        if save_fig_path is not None:
            pylab.savefig(save_fig_path)

        if not interactive and fig_show:
            pylab.show()

    def redraw(self) -> None:
        """Redraw the dendrogram."""
        if self.active:
            if not self.drawCount:
                for psection in self.psections:
                    redraw_psection(psection)
                self.canvas.blit(self.ax.bbox)
                self.drawCount = 1
            else:
                self.drawCount = self.drawCount - 1



"""Utility functions used within BlueCellulab."""

from __future__ import annotations
import contextlib
import io
import json

import numpy as np


def run_once(func):
    """A decorator to ensure a function is only called once."""

    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return func(*args, **kwargs)

    wrapper.has_run = False
    return wrapper


class CaptureOutput(list):
    def __enter__(self):
        self._stringio = io.StringIO()
        self._redirect_stdout = contextlib.redirect_stdout(self._stringio)
        self._redirect_stdout.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._redirect_stdout.__exit__(exc_type, exc_val, exc_tb)
        self.extend(self._stringio.getvalue().splitlines())


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Synapse Properties."""

from __future__ import annotations
from enum import Enum
from types import MappingProxyType
from typing import Any

from bluecellulab import BLUEPY_AVAILABLE
from bluecellulab.exceptions import ExtraDependencyMissingError

if BLUEPY_AVAILABLE:
    from bluepy.enums import Synapse as BLPSynapse


class SynapseProperty(Enum):
    PRE_GID = "pre_gid"
    AXONAL_DELAY = "axonal_delay"
    POST_SECTION_ID = "post_section_id"
    POST_SEGMENT_ID = "post_segment_id"
    POST_SEGMENT_OFFSET = "post_segment_offset"
    G_SYNX = "g_synx"
    U_SYN = "u_syn"
    D_SYN = "d_syn"
    F_SYN = "f_syn"
    DTC = "DTC"
    TYPE = "type"
    NRRP = "NRRP"
    U_HILL_COEFFICIENT = "u_hill_coefficient"
    CONDUCTANCE_RATIO = "conductance_scale_factor"
    AFFERENT_SECTION_POS = "afferent_section_pos"

    @classmethod
    def from_bluepy(cls, prop: BLPSynapse) -> SynapseProperty:
        return cls(prop.value)

    def to_bluepy(self) -> BLPSynapse:
        if not BLUEPY_AVAILABLE:
            raise ExtraDependencyMissingError("bluepy")
        return BLPSynapse(self.value)

    @classmethod
    def from_snap(cls, prop: str) -> SynapseProperty:
        return cls(snap_to_synproperty[prop])

    def to_snap(self) -> str:
        return synproperty_to_snap[self]


class SynapseProperties:
    """Synapse Properties to be retrieved from circuit and used by the
    simulator."""
    common = (
        SynapseProperty.PRE_GID,
        SynapseProperty.AXONAL_DELAY,
        SynapseProperty.POST_SECTION_ID,
        SynapseProperty.POST_SEGMENT_ID,
        SynapseProperty.POST_SEGMENT_OFFSET,
        SynapseProperty.G_SYNX,
        SynapseProperty.U_SYN,
        SynapseProperty.D_SYN,
        SynapseProperty.F_SYN,
        SynapseProperty.DTC,
        SynapseProperty.TYPE,
        SynapseProperty.NRRP,
        SynapseProperty.U_HILL_COEFFICIENT,
        SynapseProperty.CONDUCTANCE_RATIO,
        SynapseProperty.AFFERENT_SECTION_POS,
    )
    plasticity = (
        "volume_CR", "rho0_GB", "Use_d_TM", "Use_p_TM", "gmax_d_AMPA",
        "gmax_p_AMPA", "theta_d", "theta_p"
    )


snap_to_synproperty = MappingProxyType({
    "@source_node": SynapseProperty.PRE_GID,
    "delay": SynapseProperty.AXONAL_DELAY,
    "afferent_section_id": SynapseProperty.POST_SECTION_ID,
    "afferent_segment_id": SynapseProperty.POST_SEGMENT_ID,
    "afferent_segment_offset": SynapseProperty.POST_SEGMENT_OFFSET,
    "conductance": SynapseProperty.G_SYNX,
    "u_syn": SynapseProperty.U_SYN,
    "depression_time": SynapseProperty.D_SYN,
    "facilitation_time": SynapseProperty.F_SYN,
    "decay_time": SynapseProperty.DTC,
    "syn_type_id": SynapseProperty.TYPE,
    "n_rrp_vesicles": SynapseProperty.NRRP,
    "u_hill_coefficient": SynapseProperty.U_HILL_COEFFICIENT,
    "conductance_scale_factor": SynapseProperty.CONDUCTANCE_RATIO,
    "afferent_section_pos": SynapseProperty.AFFERENT_SECTION_POS,
})


# the inverse of snap_to_synproperty
synproperty_to_snap = MappingProxyType({
    v: k for k, v in snap_to_synproperty.items()
})


def properties_from_snap(
    props: list[str],
) -> list[SynapseProperty | str]:
    """Convert list of SNAP Synapse properties to SynapseProperty, spare
    'str's."""
    return [
        SynapseProperty.from_snap(prop)
        if prop in snap_to_synproperty
        else prop
        for prop in props
    ]


def properties_to_snap(props: list[SynapseProperty | str]) -> list[str]:
    """Convert a list of SynapseProperty to SNAP properties, spare 'str's."""
    return [
        prop.to_snap()
        if isinstance(prop, SynapseProperty)
        else prop
        for prop in props
    ]


def properties_from_bluepy(
    props: list[BLPSynapse | str],
) -> list[SynapseProperty | str]:
    """Convert list of bluepy Synapse properties to SynapseProperty, spare
    'str's."""
    if not BLUEPY_AVAILABLE:
        raise ExtraDependencyMissingError("bluepy")
    res: list[SynapseProperty | str] = []
    for prop in props:
        if isinstance(prop, BLPSynapse):
            res.append(SynapseProperty.from_bluepy(prop))
        elif prop == "afferent_section_pos":  # jira_url/project/issues/browse/NSETM-2313
            res.append(SynapseProperty.AFFERENT_SECTION_POS)
        else:
            res.append(prop)
    return res


def properties_to_bluepy(props: list[SynapseProperty | str]) -> list[BLPSynapse | str]:
    """Convert a list of SynapseProperty to bluepy Synapse properties, spare
    'str's."""
    # bluepy does not have AFFERENT_SECTION_POS atm.
    # jira_url/project/issues/browse/NSETM-2313
    bluepy_recognised_props = props.copy()
    removed_afferent_section_pos = False
    if SynapseProperty.AFFERENT_SECTION_POS in bluepy_recognised_props:
        removed_afferent_section_pos = True
        bluepy_recognised_props.remove(SynapseProperty.AFFERENT_SECTION_POS)
    res = [
        prop.to_bluepy()
        if isinstance(prop, SynapseProperty)
        else prop
        for prop in bluepy_recognised_props
    ]
    if removed_afferent_section_pos:
        res.append("afferent_section_pos")
    return res


def synapse_property_encoder(dct: dict[SynapseProperty | str, Any]) -> dict[str, Any]:
    """Convert SynapseProperty enum keys to strings."""
    return {key.name if isinstance(key, SynapseProperty) else key: value for key, value in dct.items()}


def synapse_property_decoder(dct: dict) -> dict[str | SynapseProperty, Any]:
    """For JSON decoding of dict containing SynapseProperty."""
    transformed_dict: dict[str | SynapseProperty, Any] = {}
    for key, value in dct.items():
        if key in SynapseProperty._member_names_:
            transformed_dict[SynapseProperty[key]] = value
        else:
            transformed_dict[key] = value
    return transformed_dict



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
project = 'BlueCelluLab'
author = 'Blue Brain Project/EPFL'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

templates_path = ['_templates']
exclude_patterns = []

autosummary_mock_imports = [  # these modules are not publicly available
    'bluepy',
    'bluepy_configfile'
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx-bluebrain-theme"
html_static_path = ['_static']

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

extensions = [
    "sphinx.ext.autodoc",  # Core library for html generation from docstrings
    "sphinx.ext.autosummary",  # Create neat summary tables
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.doctest",
    "matplotlib.sphinxext.plot_directive",
]

# Configure doctest
doctest_test_doctest_blocks = 'default'

# For matplotlib plots in Sphinx
plot_include_source = True
plot_html_show_source_link = False

autosummary_generate = True  # Turn on sphinx.ext.autosummary
autodoc_default_options = {
    "members": True,
}

# Autodoc-typehints

always_document_param_types = True
typehints_use_rtype = True
